/* Copyright 2018-2019 CS Systèmes d'Information
 * Licensed to CS Systèmes d'Information (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.web.doc.examples;

import java.io.InputStreamReader;
import java.io.StringWriter;
import java.net.URL;
import java.net.URLEncoder;

import javax.net.ssl.HttpsURLConnection;

import org.orekit.web.json.AzElPosition;
import org.orekit.web.json.CartesianPosition;
import org.orekit.web.json.Frame;
import org.orekit.web.json.GeographicPosition;
import org.orekit.web.json.Position;
import org.orekit.web.json.PvCoordinates;
import org.orekit.web.json.TopocentricFrame;
import org.orekit.web.json.tools.ConvertFrameRequest;
import org.orekit.web.json.tools.ConvertLocationRequest;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.io.CharStreams;

public class CartesianToAzelExample {

    /**
     * Main.
     * @param args oraas url
     */
	public static void main(String[] args) {

	    if (args.length < 1) {
            System.err.println("Usage: CartesianToAzEl needs ORaaS_server_url");
            System.exit(1);
        }

        // Create a trust manager that does not validate certificate chains
        try {
            AllCertificateValidator.configure();
            callORaaSConverters(args[0], false);
            
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }

	}

	private static void callORaaSConverters(String oraas, boolean print) throws Exception {
   
	    // ITRF frame
	    String itrf = "CIO/2010-based ITRF simple EOP";

	    // Point
	    GeographicPosition point = new GeographicPosition();
	    point.setLat(0.761);
	    point.setLon(0.025);
	    point.setAlt(150.0);
	    point.setEcefFrameName(itrf);

	    // Topocentric frame
	    TopocentricFrame topocentric = new TopocentricFrame();
	    topocentric.setOrigin(point);

	     // Date
        String date = "2019-03-12T00:00:00.000";

        // Input frame
        Frame inputFrame = new Frame();
        inputFrame.setFrameName("EME2000");

        // Output frame
        Frame outputFrame = new Frame();
        outputFrame.setTopocentricFrame(topocentric);

        // Time scale
        String timeScale = "UTC";

        // Cartesian position
        PvCoordinates pv = new PvCoordinates();
        pv.setPx(-4556412.23);
        pv.setPy(760212.52);
        pv.setPz(4384932.89);
        pv.setVx(762.5);
        pv.setVy(-1474.4);
        pv.setVz(55.4);
        
        // Convert Frame Request
        ConvertFrameRequest cfr = new ConvertFrameRequest();
        cfr.setDate(date);
        cfr.setTimeScale(timeScale);
        cfr.setPvCoordinates(pv);
        cfr.setInputFrame(inputFrame);
        cfr.setOutputFrame(outputFrame);

        // Serialize the request with jackson
        ObjectMapper om = new ObjectMapper();
        StringWriter sw = new StringWriter();
        om.writeValue(sw, cfr);
        String jsonStr = sw.toString();
        
        URL url = new URL(oraas + "/frame/convert-pv"
        		+ "?json="+ URLEncoder.encode(jsonStr, "UTF-8"));
		HttpsURLConnection connection = (HttpsURLConnection) url.openConnection();
		connection.setRequestMethod("GET");
		connection.setRequestProperty("Accept", "application/json");
		connection.setDoOutput(true);
        connection.connect();
        
        String body = CharStreams.toString(new InputStreamReader(connection.getInputStream()));
		
		PvCoordinates response = new ObjectMapper().readValue(body, PvCoordinates.class);

		connection.disconnect();

		// Converted position
		final Position pos = new Position();
		final CartesianPosition cartesian = new CartesianPosition();
		cartesian.setX(response.getPx());
		cartesian.setY(response.getPy());
		cartesian.setZ(response.getPz());
		cartesian.setFrame(outputFrame);
		pos.setCartesian(cartesian);

		// Convert Location Request
		ConvertLocationRequest cpr = new ConvertLocationRequest();
		cpr.setDate(date);
		cpr.setTimeScale(timeScale);
		cpr.setLocation(pos);
		cpr.setAzElOutputTopocentricFrame(topocentric);

        // Serialize the request with jackson
        ObjectMapper om2 = new ObjectMapper();
        StringWriter sw2 = new StringWriter();
        om2.writeValue(sw2, cpr);
        String jsonStr2 = sw2.toString();
        
        URL url2 = new URL(oraas + "/location/convert"
                + "?json="+ URLEncoder.encode(jsonStr2, "UTF-8"));
        HttpsURLConnection connection2 = (HttpsURLConnection) url2.openConnection();
        connection2.setRequestMethod("GET");
        connection2.setRequestProperty("Accept", "application/json");
        connection2.setDoOutput(true);
        connection2.connect();
        
        String body2 = CharStreams.toString(new InputStreamReader(connection2.getInputStream()));

        Position position2 = new ObjectMapper().readValue(body2, Position.class);
        AzElPosition azel = position2.getAzEl();

        if (print) {
            System.out.println("azimuth = " + azel.getAzimuth() + " rad");
            System.out.println("elevation = " + azel.getElevation() + " rad"); 
        }

    }

}
