/* Copyright 2018-2019 CS Systèmes d'Information
 * Licensed to CS Systèmes d'Information (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.web.doc.examples;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.HttpCookie;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.HttpsURLConnection;

import org.eclipse.jetty.websocket.client.ClientUpgradeRequest;
import org.eclipse.jetty.websocket.client.WebSocketClient;
import org.orekit.web.json.GeographicPosition;
import org.orekit.web.json.OrbitBulletin;
import org.orekit.web.json.PvCoordinates;
import org.orekit.web.json.enums.MeasurementsFileFormat;
import org.orekit.web.json.od.Estimator;
import org.orekit.web.json.od.Measurements;
import org.orekit.web.json.od.Measurements.MeasurementsFile;
import org.orekit.web.json.od.Measurements.MeasurementsStation;
import org.orekit.web.json.od.OrbitDeterminationRequest;
import org.orekit.web.json.propagation.CentralBodyGravity;
import org.orekit.web.json.propagation.ForceModel;
import org.orekit.web.json.propagation.NumericalPropagator;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.io.CharStreams;
import com.google.common.io.Files;


public class OrbitDeterminationExample {

	private static final String TEST_GUID = "odExample";
	private static final String REQUEST_PART_NAME = "w3bTdmFile";
	
	public static void main(String[] args) {

		try {
			if (args.length < 2) {
				System.err.println("Usage: OrbitDeterminationExample ORaaS_server_url W3B_TDM_File_name");
				System.exit(1);
			}
			File tdm = new File(args[1]);
			if (! tdm.exists()) {
				System.err.println("File not found:" + args[1]);
				System.exit(1);
			}
			
			// Create a trust manager that does not validate certificate chains
			AllCertificateValidator.configure();
			
			String runId = createNewRun(args[0], tdm);
			listRuns(args[0]);
			
			// connection to the web socket - wait for the result
			WebSocketClient client = new WebSocketClient();
			try {
				client.start();
				ODClientSocket runSocket = new ODClientSocket();
				URI uri = new URI(args[0].toLowerCase().replace("http", "ws") 
						+ "/od/run/progress/ws?runId=" + runId);
				ClientUpgradeRequest request = new ClientUpgradeRequest();
				
				List<HttpCookie> cookies = new ArrayList<HttpCookie>(1);
				cookies.add(new HttpCookie("guid", TEST_GUID));
				request.setCookies(cookies );
				
				client.connect(runSocket, uri, request);
				
				// wait for closed socket connection.
				runSocket.awaitClose(2 * 60, TimeUnit.SECONDS);
			
			} finally {
				client.stop();
			}
			getOrbit(args[0], runId);
//			deleteRun(args[0], runId);
			
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}
	}

	private static void getOrbit(String oraas, String runId) throws IOException {
		URL url = new URL(oraas + "/od/runs/" + runId + "/orbit");
		HttpsURLConnection connection = (HttpsURLConnection) url.openConnection();
		connection.setRequestMethod("GET");
		connection.setRequestProperty("Cookie", "guid=" + TEST_GUID); 
		connection.setRequestProperty("Accept", "application/json");
		connection.setDoOutput(true);
		connection.connect();
		System.out.println(CharStreams.toString(new InputStreamReader(connection.getInputStream())));
	}

	private static void listRuns(String oraas) throws IOException {
		URL url = new URL(oraas + "/od/runs");
		HttpsURLConnection connection = (HttpsURLConnection) url.openConnection();
		connection.setRequestMethod("GET");
		connection.setRequestProperty("Cookie", "guid=" + TEST_GUID); 
		connection.setRequestProperty("Accept", "application/json");
		connection.setDoOutput(true);
		connection.connect();
		System.out.println(CharStreams.toString(new InputStreamReader(connection.getInputStream())));
	}

	private static void deleteRun(String oraas, String id) throws IOException {
		URL url = new URL(oraas + "/od/runs/"+ id);
		HttpsURLConnection connection = (HttpsURLConnection) url.openConnection();
		connection.setRequestMethod("DELETE");
		connection.setRequestProperty("Cookie", "guid=" + TEST_GUID); 
		connection.setRequestProperty("Accept", "application/json");
		connection.setDoOutput(true);
		connection.connect();
		int status = connection.getResponseCode();
		System.out.println(status);
	}

	private static String createNewRun(String oraas, File tdm) throws Exception {
		URL url = new URL(oraas + "/od/run");

		String boundary = Long.toHexString(System.currentTimeMillis()); // Just generate some unique random value.
		String CRLF = "\r\n"; // Line separator required by multipart/form-data.
		String charset = "UTF-8";
		
		HttpsURLConnection connection = (HttpsURLConnection) url.openConnection();		
		connection.setRequestMethod("POST");
		connection.setRequestProperty("Cookie", "guid=" + TEST_GUID); 
		
		connection.setDoOutput(true);			
		connection.setRequestProperty("Content-Type", "multipart/form-data; boundary=" + boundary);

		OutputStream output = connection.getOutputStream();
	    PrintWriter writer = new PrintWriter(new OutputStreamWriter(output, charset), true);
	
	    // Build the object sent as json parameter
	    OrbitDeterminationRequest odr = new OrbitDeterminationRequest();
	    
	    // guessed orbit
	    OrbitBulletin ob = new OrbitBulletin();
	    odr.setInitialOrbit(ob);
	    ob.setEciFrameName("EME2000");
	    ob.setDate("2010-11-02T02:56:15.690");
	    ob.setTimeScale("UTC");
    	PvCoordinates c = new PvCoordinates();
		ob.setCartesian(c);
		c.setPx(-40517522.9);
		c.setPy(-10003079.9);
		c.setPz(166792.8);
		c.setVx(762.559);
		c.setVy(-1474.468);
		c.setVz(55.430);
		
		// TDM measurements file
	    Measurements measurements = new Measurements();
	    measurements.setFiles(new MeasurementsFile[1]);
	    measurements.getFiles()[0] = new MeasurementsFile();
	    measurements.getFiles()[0].setFormat(MeasurementsFileFormat.TDM_KVN.toString());
	    measurements.getFiles()[0].setRequestPartName(REQUEST_PART_NAME);

	    // Station Fucino
	    measurements.setStations(new MeasurementsStation[5]);
	    measurements.getStations()[0] = new MeasurementsStation();
	    measurements.getStations()[0].setName("Fucino");
	    measurements.getStations()[0].setGeographicPosition(new GeographicPosition());
	    measurements.getStations()[0].getGeographicPosition().setAlt(671.35);
	    measurements.getStations()[0].getGeographicPosition().setLat(41.98);
	    measurements.getStations()[0].getGeographicPosition().setLon(13.60);

        // Station Kumsan
        measurements.getStations()[1] = new MeasurementsStation();
        measurements.getStations()[1].setName("Kumsan");
        measurements.getStations()[1].setGeographicPosition(new GeographicPosition());
        measurements.getStations()[1].getGeographicPosition().setAlt(180.55);
        measurements.getStations()[1].getGeographicPosition().setLat(36.12);
        measurements.getStations()[1].getGeographicPosition().setLon(127.49);

        // Station Uralla
        measurements.getStations()[2] = new MeasurementsStation();
        measurements.getStations()[2].setName("Uralla");
        measurements.getStations()[2].setGeographicPosition(new GeographicPosition());
        measurements.getStations()[2].getGeographicPosition().setAlt(1163.27);
        measurements.getStations()[2].getGeographicPosition().setLat(-30.63);
        measurements.getStations()[2].getGeographicPosition().setLon(151.56);

        // Station Pretoria
        measurements.getStations()[3] = new MeasurementsStation();
        measurements.getStations()[3].setName("Pretoria");
        measurements.getStations()[3].setGeographicPosition(new GeographicPosition());
        measurements.getStations()[3].getGeographicPosition().setAlt(1566.63);
        measurements.getStations()[3].getGeographicPosition().setLat(-25.88);
        measurements.getStations()[3].getGeographicPosition().setLon(27.71);

        // Station Pretoria
        measurements.getStations()[4] = new MeasurementsStation();
        measurements.getStations()[4].setName("CastleRock");
        measurements.getStations()[4].setGeographicPosition(new GeographicPosition());
        measurements.getStations()[4].getGeographicPosition().setAlt(2095.38);
        measurements.getStations()[4].getGeographicPosition().setLat(39.28);
        measurements.getStations()[4].getGeographicPosition().setLon(-104.81);

        odr.setMeasurements(measurements);

	    // configuration of the algorithm (mandatory fields)
	    Estimator estimator = new Estimator();
	    estimator.setNumericalPropagator(new NumericalPropagator());
	    ForceModel forceModel = new ForceModel();
	    forceModel.setCentralBodyGravity(new CentralBodyGravity());
	    forceModel.getCentralBodyGravity().setDegree(20);
	    forceModel.getCentralBodyGravity().setOrder(20);
		estimator.getNumericalPropagator().setForceModel(forceModel);
		odr.setEstimator(estimator);

		// Serialize the request with jackson
		ObjectMapper om = new ObjectMapper();
		StringWriter sw = new StringWriter();
		om.writeValue(sw, odr);
		String jsonStr = sw.toString();
		
		// Write the multi-part request
	    writer.append("--" + boundary).append(CRLF);
	    writer.append("Content-Disposition: form-data; name=\"")
	    		.append("json").append("\"").append(CRLF);
	    writer.append("Content-Type: text/plain; charset=" + charset).append(CRLF);
		writer.append(CRLF).append(jsonStr).append(CRLF).flush();

	    // Measurement file.
	    writer.append("--" + boundary).append(CRLF);
	    writer.append("Content-Disposition: form-data; name=\"")
	    .append(REQUEST_PART_NAME)
	    .append("\"; filename=\"" + tdm.getName() + "\"").append(CRLF);
	    writer.append("Content-Type: text/plain; charset=" + charset).append(CRLF); // Text file itself must be saved in this charset!
	    writer.append(CRLF).flush();
	    Files.copy(tdm, output);
	    
	    output.flush(); // Important before continuing with writer!
	    writer.append(CRLF).flush(); // CRLF is important! It indicates end of boundary.

	    // End of multipart/form-data.
	    writer.append("--" + boundary + "--").append(CRLF).flush();
				    
		connection.connect();
	    
	    int responseCode = ((HttpsURLConnection) connection).getResponseCode();
	    assert(responseCode == 200); 
		
		String body = CharStreams.toString(new InputStreamReader(connection.getInputStream()));
		try {
			// ORaaS identifies the id by a number
			Integer.parseInt(body);
		} catch (NumberFormatException e) {
			throw new Exception("ORaaS service response: " + body);
		}		
		return body;
	}

}
